# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_utilities.overrides import overrides
from .abstract_dict import AbstractDict
from .abstract_view import AbstractView


class _SliceView(AbstractView):
    __slots__ = [
        "_start", "_stop"]

    def __init__(self, range_dict, start, stop):
        """
        Use :py:meth:`RangeDictionary.view_factory` to create views
        """
        super().__init__(range_dict)
        self._start = start
        self._stop = stop

    def __str__(self):
        return f"View with range: {self._start} to {self._stop}"

    @overrides(AbstractDict.ids)
    def ids(self):
        return range(self._start, self._stop)

    @overrides(AbstractDict.get_value)
    def get_value(self, key):
        return self._range_dict.get_list(key).get_single_value_by_slice(
            slice_start=self._start, slice_stop=self._stop)

    def update_save_iter_all_values(self, key):
        ranged_list = self._range_dict.get_list(key)
        for the_id in self.ids():
            yield ranged_list.get_value_by_id(the_id=the_id)

    @overrides(AbstractDict.iter_all_values, extend_defaults=True)
    def iter_all_values(self, key=None, update_save=False):
        if isinstance(key, str):
            if update_save:
                return self.update_save_iter_all_values(key)
            return self._range_dict.get_list(key).iter_by_slice(
                slice_start=self._start, slice_stop=self._stop)
        return self._range_dict.iter_values_by_slice(
            key=key, slice_start=self._start, slice_stop=self._stop,
            update_save=update_save)

    @overrides(AbstractDict.set_value)
    def set_value(self, key, value, use_list_as_value=False):
        self._range_dict.get_list(key).set_value_by_slice(
            slice_start=self._start, slice_stop=self._stop, value=value,
            use_list_as_value=use_list_as_value)

    @overrides(AbstractDict.iter_ranges)
    def iter_ranges(self, key=None):
        return self._range_dict.iter_ranges_by_slice(
            key=key, slice_start=self._start, slice_stop=self._stop)



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .x_view import XView
from .y_view import YView


class DoubleDict(object):
    __slots__ = [
        "_matrix", "_xtype", "_ytype"]

    def __init__(self, xtype, ytype, matrix):
        self._xtype = xtype
        self._ytype = ytype
        self._matrix = matrix

    def __getitem__(self, key):
        if isinstance(key, self._xtype):
            return XView(x=key, matrix=self._matrix)
        if isinstance(key, self._ytype):
            return YView(y=key, matrix=self._matrix)
        raise KeyError(f"Key {key} has an unexpected type")

    def __setitem__(self, key, value):
        try:
            if isinstance(key, self._xtype):
                correct = all(isinstance(y, self._ytype) for y in value.keys())
                if not correct:
                    raise ValueError(
                        f"All keys in the value must be of type {self._ytype}")
                for y in value.keys():
                    self._matrix.set_data(x=key, y=y, value=value[y])
            elif isinstance(key, self._ytype):
                correct = all(isinstance(x, self._xtype) for x in value.keys())
                if not correct:
                    raise ValueError(
                        f"All keys in the value must be of type {self._xtype}")
                for x in value.keys():
                    self._matrix.set_data(x=x, y=key, value=value[x])
            else:
                raise KeyError(f"Key {key} has an unexpected type")
        except AttributeError as e:
            raise ValueError("Value must of type dict. Or at least "
                             "implement keys() and __getitem__") from e



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import logging
import sys
import numpy

logger = logging.getLogger(__file__)


class AbstractSized(object):
    """
    Base class for slice and ID checking against size.
    Subclasses of this support the `len()` built-in.
    """

    __slots__ = (
        "_size")

    def __init__(self, size):
        """
        :param int size: Fixed length of the list.
        """
        # Strictly doesn't need to be int, but really should be!
        self._size = max(int(round(size)), 0)

    def __len__(self):
        """
        Size of the list, irrespective of actual values.

        :return: the initial and Fixed size of the list
        """
        return self._size

    @staticmethod
    def _is_id_type(the_id):
        """
        Check if the given ID has a type acceptable for IDs.
        """
        return isinstance(the_id, int)

    def _check_id_in_range(self, the_id):
        if the_id < 0:
            if self._is_id_type(the_id):
                raise IndexError(f"The index {the_id} is out of range.")
            # pragma: no cover
            raise TypeError(f"Invalid argument type {type(the_id)}.")
        if the_id >= self._size:
            if self._is_id_type(the_id):
                raise IndexError(f"The index {the_id} is out of range.")
            raise TypeError(f"Invalid argument type {type(the_id)}.")

    def _check_slice_in_range(self, slice_start, slice_stop):
        if slice_start is None:
            slice_start = 0
        elif slice_start < 0:
            slice_start = self._size + slice_start
            if slice_start < 0:
                if not self._is_id_type(slice_start):
                    raise TypeError(
                        f"Invalid argument type {type(slice_start)}.")
                logger.warning(
                    "Specified slice start was %d while size is only %d. "
                    "Therefore slice will start at index 0",
                    slice_start - self._size, self._size)
                slice_start = 0
        elif slice_start >= len(self):
            logger.warning(
                "Specified slice start was %d while size is only %d. "
                "Therefore slice will be empty",
                slice_start - self._size, self._size)
            return (self._size, self._size)

        if slice_stop is None or slice_stop == sys.maxsize:
            slice_stop = self._size
        elif slice_stop < 0:
            slice_stop = self._size + slice_stop

        if slice_start > slice_stop:
            if not self._is_id_type(slice_start):
                raise TypeError(f"Invalid argument type {type(slice_start)}.")
            if not self._is_id_type(slice_stop):
                raise TypeError(f"Invalid argument type {type(slice_stop)}.")
            logger.warning(
                "Specified slice has a start %d greater than its stop %d "
                "(based on size %d). Therefore slice will be empty",
                slice_start, slice_stop, self._size)
            return (self._size, self._size)
        if slice_stop > len(self):
            if not self._is_id_type(slice_stop):
                raise TypeError(f"Invalid argument type {type(slice_stop)}.")
            logger.warning(
                "Specified slice has a start %d equal to its stop %d "
                "(based on size %d). Therefore slice will be empty",
                slice_start, slice_stop, self._size)
        if slice_stop < 0:
            logger.warning(
                "Specified slice stop was %d while size is only %d. "
                "Therefore slice will be empty",
                slice_stop - self._size, self._size)
            return (self._size, self._size)
        elif slice_start > slice_stop:
            logger.warning(
                "Specified slice has a start %d greater than its stop %d "
                "(based on size %d). Therefore slice will be empty",
                slice_start, slice_stop, self._size)
            return (self._size, self._size)
        elif slice_start == slice_stop:
            logger.warning(
                "Specified slice has a start %d equal to its stop %d "
                "(based on size %d). Therefore slice will be empty",
                slice_start, slice_stop, self._size)
        elif slice_stop > len(self):
            logger.warning(
                "Specified slice stop was %d while size is only %d. "
                "Therefore slice will be truncated",
                slice_stop, self._size)
            slice_stop = self._size
        return slice_start, slice_stop

    def _check_mask_size(self, selector):
        if len(selector) < self._size:
            logger.warning(
                "The boolean mask is too short. The expected length was %d "
                "but the length was only %d. All the missing entries will be "
                "treated as False!", self._size, len(selector))
        elif len(selector) > self._size:
            logger.warning(
                "The boolean mask is too long. The expected length was %d "
                "but the length was only %d. All the missing entries will be "
                "ignored!", self._size, len(selector))

    def selector_to_ids(self, selector, warn=False):
        """
        Gets the list of IDs covered by this selector.
        The types of selector currently supported are:

        `None`:
            Returns all IDs.

        :py:class:`slice`: Standard python slice.
            Negative values and values larger than size are handled using
            slice's `indices` method.
            This could result in am empty list.

        :py:class:`int`: Handles negative values as normal.
            Checks if ID is within expected range.

        iterator(:py:class:`bool`): Used as a mask.
            If the length of the mask is longer or shorted than number of IDs
            the result is the shorter of the two,
            with the remainder of the longer ignored.

        iterator(:py:class:`int`) but not bool:
            Every value checked that it is with the range 0 to size.
            Negative values are *not* allowed.
            Original order and duplication is respected so result may be
            unordered and contain duplicates.

        :param selector: Some object that identifies a range of IDs.
        :param bool warn:
            If True, this method will warn about problems with the selector.
        :return: a (possibly sorted) list of IDs
        """
        # Check selector is an iterable using pythonic try
        try:
            iterator = iter(selector)
        except TypeError:
            iterator = None

        if iterator is not None:
            # bool is superclass of int so if any are bools all must be
            if any(isinstance(item, (bool, numpy.bool_)) for item in selector):
                if all(isinstance(item, (bool, numpy.bool_))
                       for item in selector):
                    if warn:
                        self._check_mask_size(selector)
                    return list(itertools.compress(
                        range(self._size), selector))
                raise TypeError(
                    "An iterable type must be all ints or all bools")
            elif all(isinstance(item, (int, numpy.integer))
                     for item in selector):
                # list converts any specific numpy types
                ids = list(selector)
                for _id in ids:
                    if _id < 0:
                        raise TypeError(
                            f"Selector includes the ID {_id} which is "
                            "less than zero")
                    if _id >= self._size:
                        raise TypeError(
                            f"Selector includes the ID {_id} which not "
                            f"less than the size {self._size}")
                return ids
            else:
                raise TypeError(
                    "An iterable type must be all ints or all bools")

        # OK lets try for None, int and slice after all
        if selector is None:
            if warn:
                logger.warning("None selector taken as all IDs")
            return range(self._size)

        if isinstance(selector, slice):
            (slice_start, slice_stop, step) = selector.indices(self._size)
            return range(slice_start, slice_stop, step)

        if isinstance(selector, int):
            if selector < 0:
                selector = self._size + selector
            if selector < 0 or selector >= self._size:
                raise TypeError(
                    f"Selector {selector-self._size} is unsupported "
                    f"for size {self._size}")
            return [selector]

        raise TypeError(f"Unexpected selector type {type(selector)}")



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import spinn_utilities.conf_loader as conf_loader
from spinn_utilities.configs import CamelCaseConfigParser
from spinn_utilities.exceptions import ConfigException
from spinn_utilities.log import FormatAdapter

# pylint: disable=global-statement
logger = FormatAdapter(logging.getLogger(__file__))

__config = None
__default_config_files = []
__config_file = None
__unittest_mode = False


def add_default_cfg(default):
    """
    Adds an extra default configuration file to be read after earlier ones.

    :param str default: Absolute path to the configuration file
    """
    if default not in __default_config_files:
        __default_config_files.append(default)


def clear_cfg_files(unittest_mode):
    """
    Clears any previous set configurations and configuration files.

    After this method :py:func:`add_default_cfg` and :py:func:`set_cfg_files`
    need to be called.

    :param bool unittest_mode: Flag to put the holder into unit testing mode
    """
    global __config, __config_file, __unittest_mode
    __config = None
    __default_config_files.clear()
    __config_file = None
    __unittest_mode = unittest_mode


def set_cfg_files(config_file, default):
    """
    Adds the configuration files to be loaded.

    :param str config_file:
        The base name of the configuration file(s).
        Should not include any path components.
    :param str default:
        Full path to the extra file to get default configurations from.
    """
    global __config_file
    __config_file = config_file
    add_default_cfg(default)


def _pre_load_config():
    """
    Loads configurations due to early access to a configuration value.

    :raises ConfigException: Raise if called before setup
    """
    # If you getthis error during a unittest then unittest_step was not called
    if not __unittest_mode:
        raise ConfigException(
            "Accessing config values before setup is not supported")
    load_config()


def load_config():
    """
    Reads in all the configuration files, resetting all values.

    :raises ConfigException: If called before setting defaults
    """
    global __config
    if not __default_config_files:
        raise ConfigException("No default configs set")
    if __config_file:
        __config = conf_loader.load_config(
            filename=__config_file, defaults=__default_config_files)
    else:
        __config = CamelCaseConfigParser()
        for default in __default_config_files:
            __config.read(default)


def get_config_str(section, option):
    """
    Get the string value of a configuration option.

    :param str section: What section to get the option from.
    :param str option: What option to read.
    :return: The option value
    :rtype: str or None
    """
    try:
        return __config.get_str(section, option)
    except AttributeError:
        pass
    _pre_load_config()
    return __config.get_str(section, option)


def get_config_str_list(section, option, token=","):
    """
    Get the string value of a configuration option split into a list.

    :param str section: What section to get the option from.
    :param str option: What option to read.
    :param token: The token to split the string into a list
    :return: The list (possibly empty) of the option values
    :rtype: list(str)
    """
    try:
        return __config.get_str_list(section, option, token)
    except AttributeError:
        pass
    _pre_load_config()
    return __config.get_str_list(section, option, token)


def get_config_int(section, option):
    """
    Get the integer value of a configuration option.

    :param str section: What section to get the option from.
    :param str option: What option to read.
    :return: The option value
    :rtype: int
    """
    try:
        return __config.get_int(section, option)
    except AttributeError:
        pass
    _pre_load_config()
    return __config.get_int(section, option)


def get_config_float(section, option):
    """
    Get the float value of a configuration option.

    :param str section: What section to get the option from.
    :param str option: What option to read.
    :return: The option value.
    :rtype: float
    """
    try:
        return __config.get_float(section, option)
    except AttributeError:
        pass
    _pre_load_config()
    return __config.get_float(section, option)


def get_config_bool(section, option):
    """
    Get the boolean value of a configuration option.

    :param str section: What section to get the option from.
    :param str option: What option to read.
    :return: The option value.
    :rtype: bool
    """
    try:
        return __config.get_bool(section, option)
    except AttributeError:
        pass
    _pre_load_config()
    return __config.get_bool(section, option)


def set_config(section, option, value):
    """
    Sets the value of a configuration option.

    This method should only be called by the simulator or by unit tests.

    :param str section: What section to set the option in.
    :param str option: What option to set.
    :param object value: Value to set option to
    :raises ConfigException: If called unexpectedly
    """
    if __config is None:
        if __unittest_mode:
            load_config()
        else:
            # The actual error is that load_config should be called before
            # set_config but this discourages the use outside of unittests
            raise ConfigException(
                "set_config should only be called by unittests "
                "which should have called unittest_setup")
    __config.set(section, option, value)
    # Intentionally no try here to force tests that set to
    # load_default_configs before AND after


def has_config_option(section, option):
    """
    Check if the section has this configuration option.

    :param str section: What section to check
    :param str option: What option to check.
    :rtype: bool
    :return: True if and only if the option is defined. It may be `None`
    """
    try:
        return __config.has_option(section, option)
    except AttributeError:
        pass
    _pre_load_config()
    return __config.has_option(section, option)


def config_options(section):
    """
    Return a list of option names for the given section name.

    :param str section: What section to list options for.
    """
    return __config.options(section)


def _check_lines(py_path, line, lines, index, method):
    """
    Support for `_check_python_file`. Gets section and option name.

    :param str line: Line with get_config call
    :param list(str) lines: All lines in the file
    :param int index: index of line with `get_config` call
    :raises ConfigException: If an unexpected or uncovered `get_config` found
    """
    while ")" not in line:
        index += 1
        line += lines[index]
    parts = line[line.find("(", line.find("get_config")) + 1:
                 line.find(")")].split(",")
    section = parts[0].strip().replace("'", "").replace('"', '')
    option = parts[1].strip()
    if option[0] == "'":
        option = option.replace("'", "")
    elif option[0] == '"':
        option = option.replace('"', '')
    else:
        print(line)
        return
    try:
        method(section, option)
    except Exception as original:
        raise ConfigException(
            f"failed in line:{index} of file: {py_path} with "
            f"section:{section} option:{option}") from original


def _check_python_file(py_path):
    """
    A testing function to check that all the `get_config` calls work.

    :param str py_path: path to file to be checked
    :raises ConfigException: If an unexpected or uncovered `get_config` found
    """
    with open(py_path, 'r', encoding="utf-8") as py_file:
        lines = py_file.readlines()
        for index, line in enumerate(lines):
            if "get_config_bool(" in line:
                _check_lines(py_path, line, lines, index, get_config_bool)
            if "get_config_float(" in line:
                _check_lines(py_path, line, lines, index, get_config_float)
            if "get_config_int(" in line:
                _check_lines(py_path, line, lines, index, get_config_int)
            if "get_config_str(" in line:
                _check_lines(py_path, line, lines, index, get_config_str)
            if "get_config_str_list(" in line:
                _check_lines(py_path, line, lines, index, get_config_str_list)


def _check_python_files(directory):
    for root, _, files in os.walk(directory):
        for file_name in files:
            if file_name.endswith(".py"):
                py_path = os.path.join(root, file_name)
                _check_python_file(py_path)


def _find_double_defaults(repeaters=None):
    """
    Testing function to identify any configuration options in multiple default
    files.

    :param repeaters: List of options that are expected to be repeated.
    :type repeaters: None or list(str)
    :raises ConfigException:
        If two defaults configuration files set the same value
    """
    config1 = CamelCaseConfigParser()
    for default in __default_config_files[:-1]:
        config1.read(default)
    config2 = CamelCaseConfigParser()
    config2.read(__default_config_files[-1])
    if repeaters is None:
        repeaters = []
    else:
        repeaters = map(config2.optionxform, repeaters)
    for section in config2.sections():
        for option in config2.options(section):
            if config1.has_option(section, option):
                if option not in repeaters:
                    raise ConfigException(
                        f"cfg:{__default_config_files[-1]} "
                        f"repeats [{section}]{option}")


def _check_cfg_file(config1, cfg_path):
    """
    Support method for :py:func:`check_cfgs`.

    :param CamelCaseConfigParser config1:
    :param str cfg_path:
    :raises ConfigException: If an unexpected option is found
    """
    config2 = CamelCaseConfigParser()
    config2.read(cfg_path)
    for section in config2.sections():
        if not config1.has_section(section):
            raise ConfigException(
                f"cfg:{cfg_path} has unexpected section [{section}]")
        for option in config2.options(section):
            if not config1.has_option(section, option):
                raise ConfigException(
                    f"cfg:{cfg_path} "
                    f"has unexpected options [{section}]{option}")


def _check_cfgs(path):
    """
    A testing function check local configuration files against the defaults.

    It only checks that the option exists in a default.
    It does not check if the option is used or if the value is the expected
    type.

    :param str path: Absolute path to the parent directory to search
    :raises ConfigException: If an unexpected option is found
    """
    config1 = CamelCaseConfigParser()
    for default in __default_config_files:
        config1.read(default)
    directory = os.path.dirname(path)
    for root, _, files in os.walk(directory):
        for file_name in files:
            if file_name.endswith(".cfg"):
                cfg_path = os.path.join(root, file_name)
                if cfg_path in __default_config_files:
                    continue
                print(cfg_path)
                _check_cfg_file(config1, cfg_path)


def run_config_checks(directories, *, exceptions=None, repeaters=None):
    """
    Master test.

    :param module:
    :param exceptions:
    :param repeaters:
    :raises ConfigException: If an incorrect directory passed in
    """
    if isinstance(directories, str):
        directories = [directories]

    if exceptions is None:
        exceptions = []
    elif isinstance(exceptions, str):
        exceptions = [exceptions]

    _find_double_defaults(repeaters)

    config1 = CamelCaseConfigParser()
    config1.read(__default_config_files)

    for directory in directories:
        if not os.path.isdir(directory):
            raise ConfigException(f"Unable find {directory}")
        for root, _, files in os.walk(directory):
            for file_name in files:
                if file_name in exceptions:
                    pass
                elif file_name.endswith(".cfg"):
                    cfg_path = os.path.join(root, file_name)
                    if cfg_path in __default_config_files:
                        continue
                    print(cfg_path)
                    _check_cfg_file(config1, cfg_path)
                elif file_name.endswith(".py"):
                    py_path = os.path.join(root, file_name)
                    _check_python_file(py_path)



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_utilities.abstract_base import AbstractBase, abstractmethod


class AbstractMatrix(object, metaclass=AbstractBase):
    """
    A rectangular 2D collection of data.
    """
    __slots__ = []

    @abstractmethod
    def get_data(self, x, y):
        """
        Get the value at a particular X,Y coordinate.
        """

    @abstractmethod
    def set_data(self, x, y, value):
        """
        Set the value at a particular X,Y coordinate.
        """



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .camel_case_config_parser import CamelCaseConfigParser
from .case_sensitive_parser import CaseSensitiveParser
from .config_template_exception import ConfigTemplateException
from .no_config_found_exception import NoConfigFoundException
from .unexpected_config_exception import UnexpectedConfigException

__all__ = [
    "CamelCaseConfigParser", "CaseSensitiveParser",
    "ConfigTemplateException", "NoConfigFoundException",
    "UnexpectedConfigException"]



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
from spinn_utilities.overrides import overrides
from .abstract_matrix import AbstractMatrix


class DemoMatrix(AbstractMatrix):
    __slots__ = [
        "data"]

    def __init__(self):
        self.data = defaultdict(dict)

    @overrides(AbstractMatrix.get_data)
    def get_data(self, x, y):
        return self.data[x][y]

    @overrides(AbstractMatrix.set_data)
    def set_data(self, x, y, value):
        self.data[x][y] = value



# Copyright (c) 2021 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum


class ResetStatus(Enum):
    """
    Different states the reset could be in.

    This class is design to used internally by UtilsDataView.
    """
    NOT_SETUP = (0)
    SETUP = (1)
    HAS_RUN = (2)
    SOFT_RESET = (3)
    HARD_RESET = (4)

    def __new__(cls, value):
        # pylint: disable=protected-access
        obj = object.__new__(cls)
        obj._value_ = value
        return obj



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .abstract_base import AbstractBase, abstractmethod


class AbstractContextManager(object, metaclass=AbstractBase):
    """
    Closable class that supports being used as a simple context manager.
    """

    __slots__ = []

    @abstractmethod
    def close(self):
        """
        How to actually close the underlying resources.
        """

    def _context_entered(self):
        """
        Called when the context is entered. The result is ignored.
        """

    def _context_exception_occurred(self, exc_type, exc_val, exc_tb):
        """
        Called when an exception occurs during the `with` context, *after*
        the context has been closed.

        :param type exc_type:
        :param object exc_val:
        :param traceback exc_tb:
        """

    def __enter__(self):
        self._context_entered()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        if exc_type:
            self._context_exception_occurred(exc_type, exc_val, exc_tb)
        return False



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
A trimmed down version of standard Python Abstract Base classes.
"""


def abstractmethod(funcobj):
    """
    A decorator indicating abstract methods.

    Requires that the metaclass is :py:class:`AbstractBase` or derived from
    it. A class that has a metaclass derived from :py:class:`AbstractBase`
    cannot be instantiated unless all of its abstract methods are overridden.
    The abstract methods can be called using any of the normal
    'super' call mechanisms.

    Usage::

        class C(object, metaclass=AbstractBase):
            @abstractmethod
            def my_abstract_method(self, ...):
                ...
    """
    funcobj.__isabstractmethod__ = True
    return funcobj


class abstractproperty(property):
    """
    A decorator indicating abstract properties.

    Requires that the metaclass is :py:class:`AbstractBase` or derived from
    it. A class that has a metaclass derived from :py:class:`AbstractBase`
    cannot be instantiated unless all of its abstract properties are
    overridden. The abstract properties can be called using any of the normal
    'super' call mechanisms.

    Usage::

        class C(object, metaclass=AbstractBase):
            @abstractproperty
            def my_abstract_property(self):
                ...

    This defines a read-only property; you can also define a read-write
    abstract property using the 'long' form of property declaration::

        class C(object, metaclass=AbstractBase):
            def getx(self): ...
            def setx(self, value): ...
            x = abstractproperty(getx, setx)

    .. note::
        When documenting abstract properties, remember to document them as if
        they are nouns, not verbs; they are things about the object that may
        be observed as many times as the user of the class desires.

    .. warning::
        Implementations should be idempotent; fetching the property twice in a
        row should get an equivalent value with no (meaningful) change to the
        state of the object (assuming no other non-property methods of the
        object are invoked between).

        This is an assumption that debuggers make. *Do not violate it!*
    """
    __isabstractmethod__ = True


class AbstractBase(type):
    """
    Metaclass for defining Abstract Base Classes (AbstractBases).

    Use this metaclass to create an AbstractBase. An AbstractBase can be
    subclassed directly, and then acts as a mix-in class.

    This is a trimmed down version of ABC.
    Unlike ABC you can not register unrelated concrete classes.

    Usage::

        class C(object, metaclass=AbstractBase):
            @abstractmethod
            def my_abstract_method(self, ...):
                ...
    """

    def __new__(cls, name, bases, namespace, **kwargs):
        # Actually make the class
        abs_cls = super().__new__(cls, name, bases, namespace, **kwargs)

        # Get set of abstract methods from namespace
        abstracts = set(nm for nm, val in namespace.items()
                        if getattr(val, "__isabstractmethod__", False))

        # Augment with abstract methods from superclasses
        for base in bases:
            for nm in getattr(base, "__abstractmethods__", set()):
                val = getattr(abs_cls, nm, None)
                if getattr(val, "__isabstractmethod__", False):
                    abstracts.add(nm)

        # Lock down the set
        abs_cls.__abstractmethods__ = frozenset(abstracts)
        return abs_cls



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class XView(object):
    """
    A view along a particular x-slice of a 2D matrix.
    """
    __slots__ = (
        "_matrix", "_x")

    def __init__(self, x, matrix):
        self._x = x
        self._matrix = matrix

    def __getitem__(self, key):
        return self._matrix.get_data(x=self._x, y=key)

    def __setitem__(self, key, value):
        self._matrix.set_data(x=self._x, y=key, value=value)



# Copyright (c) 2018 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
from spinn_utilities.overrides import overrides
_introspector = inspect.getfullargspec


class see(overrides):
    """
    A decorator for indicating that the documentation of the method
    is provided by another method with exactly the same arguments.

    .. note::
        This has the same effect as `overrides` in reality, but is provided to
        show that the method doesn't actually override.
    """

    def __init__(
            self, documentation_method, extend_doc=True,
            additional_arguments=None, extend_defaults=False):
        """
        :param documentation_method:
            The other method to reference.
        :param bool extend_doc:
            True the method doc string should be appended to the other method's
            doc string, False if the documentation should be set to the
            other method's doc string only if there isn't a doc string already
        :param iterable(str) additional_arguments:
            Additional arguments taken by the decorated method over the
            referenced method, e.g., that are to be injected
        :param bool extend_defaults:
            Whether the decorated method may specify extra defaults for the
            parameters
        """
        super().__init__(
            documentation_method, extend_doc=extend_doc,
            additional_arguments=additional_arguments,
            extend_defaults=extend_defaults)

        # Same as overrides, except name doesn't have to match
        self._relax_name_check = True

        # Give the errors a better name
        self._override_name = "documentation method"
